use std::iter;

use color_space::{Hsv, Rgb};
use either::Either;
use serde::{Deserialize, Serialize};

use super::{
    config::{DisplayConfig, LedStripConfig},
    led_color::LedColor,
};

type Point = (usize, usize);
type LedSamplePoints = Vec<Point>;

#[derive(Clone, Serialize, Deserialize, Debug)]
struct ScreenSamplePoints {
    pub top: Vec<LedSamplePoints>,
    pub bottom: Vec<LedSamplePoints>,
    pub left: Vec<LedSamplePoints>,
    pub right: Vec<LedSamplePoints>,
}

#[derive(Clone, Serialize, Deserialize, Debug)]
pub struct Screenshot {
    bitmap: Vec<u8>,
    config: DisplayConfig,
    sample_points: ScreenSamplePoints,
}

impl Screenshot {
    pub fn new(bitmap: Vec<u8>, config: DisplayConfig) -> Self {
        Self {
            bitmap,
            config,
            sample_points: Self::get_sample_points(config),
        }
    }

    fn get_sample_points(config: DisplayConfig) -> ScreenSamplePoints {
        let top = match config.top_led_strip {
            Some(led_strip_config) => Self::get_one_edge_sample_points(
                config.display_height / 8,
                config.display_width,
                led_strip_config
                    .global_start_position
                    .abs_diff(led_strip_config.global_end_position),
                5,
            ),
            None => {
                vec![]
            }
        };

        let bottom: Vec<LedSamplePoints> = match config.bottom_led_strip {
            Some(led_strip_config) => {
                let points = Self::get_one_edge_sample_points(
                    config.display_height / 9,
                    config.display_width,
                    led_strip_config
                        .global_start_position
                        .abs_diff(led_strip_config.global_end_position),
                    5,
                );
                points
                    .into_iter()
                    .map(|groups| -> Vec<Point> {
                        groups
                            .into_iter()
                            .map(|(x, y)| (x, config.display_height - y))
                            .collect()
                    })
                    .collect()
            }
            None => {
                vec![]
            }
        };

        let left: Vec<LedSamplePoints> = match config.left_led_strip {
            Some(led_strip_config) => {
                let points = Self::get_one_edge_sample_points(
                    config.display_width / 16,
                    config.display_height,
                    led_strip_config
                        .global_start_position
                        .abs_diff(led_strip_config.global_end_position),
                    5,
                );
                points
                    .into_iter()
                    .map(|groups| -> Vec<Point> {
                        groups.into_iter().map(|(x, y)| (y, x)).collect()
                    })
                    .collect()
            }
            None => {
                vec![]
            }
        };

        let right: Vec<LedSamplePoints> = match config.right_led_strip {
            Some(led_strip_config) => {
                let points = Self::get_one_edge_sample_points(
                    config.display_width / 16,
                    config.display_height,
                    led_strip_config
                        .global_start_position
                        .abs_diff(led_strip_config.global_end_position),
                    5,
                );
                points
                    .into_iter()
                    .map(|groups| -> Vec<Point> {
                        groups
                            .into_iter()
                            .map(|(x, y)| (config.display_width - y, x))
                            .collect()
                    })
                    .collect()
            }
            None => {
                vec![]
            }
        };

        ScreenSamplePoints {
            top,
            bottom,
            left,
            right,
        }
    }

    fn get_one_edge_sample_points(
        width: usize,
        length: usize,
        leds: usize,
        single_axis_points: usize,
    ) -> Vec<LedSamplePoints> {
        let cell_size_x = length as f64 / single_axis_points as f64 / leds as f64;
        let cell_size_y = width / single_axis_points;

        let point_start_y = cell_size_y / 2;
        let point_start_x = cell_size_x / 2.0;
        let point_y_list: Vec<usize> = (point_start_y..width).step_by(cell_size_y).collect();
        let point_x_list: Vec<usize> = iter::successors(Some(point_start_x), |i| {
            let next = i + cell_size_x;
            (next < (length as f64)).then_some(next)
        })
        .map(|i| i as usize)
        .collect();

        let points: Vec<Point> = point_x_list
            .iter()
            .map(|&x| point_y_list.iter().map(move |&y| (x, y)))
            .flatten()
            .collect();
        points
            .chunks(single_axis_points * single_axis_points)
            .into_iter()
            .map(|points| Vec::from(points))
            .collect()
    }

    pub fn get_colors(&self) -> DisplayColorsOfLedStrips {
        let top = self
            .get_one_edge_colors(&self.sample_points.top)
            .into_iter()
            .flat_map(|color| color.get_rgb())
            .collect();
        let bottom = self
            .get_one_edge_colors(&self.sample_points.bottom)
            .into_iter()
            .flat_map(|color| color.get_rgb())
            .collect();
        let left = self
            .get_one_edge_colors(&self.sample_points.left)
            .into_iter()
            .flat_map(|color| color.get_rgb())
            .collect();
        let right = self
            .get_one_edge_colors(&self.sample_points.right)
            .into_iter()
            .flat_map(|color| color.get_rgb())
            .collect();
        DisplayColorsOfLedStrips {
            top,
            bottom,
            left,
            right,
        }
    }

    pub fn get_one_edge_colors(
        &self,
        sample_points_of_leds: &Vec<LedSamplePoints>,
    ) -> Vec<LedColor> {
        let mut colors = vec![];
        for led_points in sample_points_of_leds {
            let mut r = 0.0;
            let mut g = 0.0;
            let mut b = 0.0;
            let len = led_points.len() as f64;
            for (x, y) in led_points {
                let position = (x + y * self.config.display_width) * 4;
                r += self.bitmap[position + 2] as f64;
                g += self.bitmap[position + 1] as f64;
                b += self.bitmap[position] as f64;
            }
            let color = LedColor::new((r / len) as u8, (g / len) as u8, (b / len) as u8);
            // paris::info!("color: {:?}", color.get_rgb());
            colors.push(color);
        }
        colors
    }

    pub fn get_top_colors(&self) -> Vec<LedColor> {
        self.get_one_edge_colors(&self.sample_points.top)
    }

    pub fn get_top_of_led_start_at(&self) -> usize {
        self.config
            .top_led_strip
            .and_then(|c| Some(c.global_start_position))
            .unwrap_or(0)
    }
    pub fn get_top_of_led_end_at(&self) -> usize {
        self.config
            .top_led_strip
            .and_then(|c| Some(c.global_end_position))
            .unwrap_or(0)
    }

    pub async fn to_webp_base64(&self) -> String {
        let bitmap = &self.bitmap;
        let mut bitflipped =
            Vec::with_capacity(self.config.display_width * self.config.display_height * 3);
        let stride = bitmap.len() / self.config.display_height;

        for y in 0..self.config.display_height {
            for x in 0..self.config.display_width {
                let i = stride * y + 4 * x;
                bitflipped.extend_from_slice(&[bitmap[i + 2], bitmap[i + 1], bitmap[i]]);
            }
        }

        let webp_memory = webp::Encoder::from_rgb(
            bitflipped.as_slice(),
            self.config.display_width as u32,
            self.config.display_height as u32,
        )
        .encode(100.0);
        return base64::encode(&*webp_memory);
    }

    pub async fn to_dto(&self) -> ScreenshotDto {
        let encode_image = self.to_webp_base64().await;
        let config = self.config.clone();
        let colors = self.get_colors();
        ScreenshotDto {
            encode_image,
            config,
            colors,
        }
    }
}

#[derive(Clone, Serialize, Deserialize, Debug)]
pub struct ScreenshotDto {
    pub config: DisplayConfig,
    pub encode_image: String,
    pub colors: DisplayColorsOfLedStrips,
}

#[derive(Clone, Serialize, Deserialize, Debug)]
pub struct DisplayColorsOfLedStrips {
    pub top: Vec<u8>,
    pub bottom: Vec<u8>,
    pub left: Vec<u8>,
    pub right: Vec<u8>,
}
